#include "global.h"
#include "mapper.h"
#include "file.h"
#include "settings.h"
#include "sample.h"
#include "state.h"
#include "scc.h"
#include "z80.h"
#include "crystal.h"
#include "sound.h"
#include "psg.h"
#include "netplay.h"
#include "am29f040b.h"
#include "io.h"
#include "tape.h"

/*

Bugs happening on a real MSX too:
 - Yie Ar Kung Fu (not the [a1] probably bugfixed version) may crash in 2 player mode vs Lan-Fang

*/

#define RAMINIT		0xff			/* value RAM is filled with on power on. alternating strings of 00s and $ffs on my MSXes */
#define UNMAPPED	0xff			/* unmapped read, $ff on my Philips VG-8235, open-bus (last opcode byte) on my Canon V-20 */

static int pslot=0;				/* ppi primary slot register */

#define DEFAULT_BIOS "cbios_main_msx1.rom"

static fp_mapread mapread_std[4][8];		/* 4 slots standard handlers */
static fp_mapwrite mapwrite_std[4][8];

static char* fn_default_bios;
static char* fn_bios=NULL; static char* n_bios=NULL;
static u32 bioscrc; static u32 default_bioscrc;
static int autodetect_crc;

static u8 ram[0x10000];				/* 64KB */
static u8 bios[0x10000];			/* 64KB, normally 32KB, but room for more (eg. cbios + logo) */
static u8 default_bios[0x8000];			/* 32KB, default C-BIOS */
static u8 dummy[0x2000];			/* 8KB filled with $ff */
static u8* tempfile[ROM_MAXPAGES];

static u8* cart[2][ROM_MAXPAGES+3];		/* max 4MB (+2 padding +1 dummy) */
static u8* cartmap[2][8];			/* cart memmaps */
static int cartbank[2][8];			/* cart banknumbers */
static u8* cartsram[2]={NULL,NULL};		/* sram */
static u8* cartsramcopy[2]={NULL,NULL};		/* sram copy */
static int cartsramsize[2]={0,0};		/* sram mem size */
static char* cartfile[2]={NULL,NULL};		/* path+filename */
static u32 cartcrc[2]={0,0};			/* checksum */
static int carttype[2];				/* romtype */
static int cartsize[2];				/* file size */
static int cartpages[2];			/* number of pages */
static int cartmask[2];				/* and mask for megaroms */

/* specific save/load state */
typedef int __fastcall (*_mapgetstatesize)(int);
static int __fastcall mapgetstatesize_nothing(int slot) { return 0; }
static _mapgetstatesize mapgetstatesize[2]={mapgetstatesize_nothing,mapgetstatesize_nothing};

typedef void __fastcall (*_mapsavestate)(int,u8**);
static void __fastcall mapsavestate_nothing(int slot,u8** s) { ; }
static _mapsavestate mapsavestate[2]={mapsavestate_nothing,mapsavestate_nothing};

typedef void __fastcall (*_maploadstatecur)(int,u8**);
static void __fastcall maploadstatecur_nothing(int slot,u8** s) { ; }
static _maploadstatecur maploadstatecur[2]={maploadstatecur_nothing,maploadstatecur_nothing};

typedef int __fastcall (*_maploadstate)(int,int,u8**);
static int __fastcall maploadstate_nothing(int slot,int v,u8** s) { return TRUE; }
static _maploadstate maploadstate[2]={maploadstate_nothing,maploadstate_nothing};

/* specific new/end frame */
typedef void (*_mapnf)(int);
static void mapnf_nothing(int slot) { ; }
static _mapnf mapnf[2]={mapnf_nothing,mapnf_nothing};
void mapper_new_frame(void) { mapnf[0](0); mapnf[1](1); }

typedef void (*_mapef)(int);
static void mapef_nothing(int slot) { ; }
static _mapef mapef[2]={mapef_nothing,mapef_nothing};
void mapper_end_frame(void) { mapef[0](0); mapef[1](1); }

/* specific sound set buffercounter */
typedef void (*_mapsetbc)(int,int);
static void mapsetbc_nothing(int slot,int bc) { ; }
static _mapsetbc mapsetbc[2]={mapsetbc_nothing,mapsetbc_nothing};
void mapper_set_buffer_counter(int bc) { mapsetbc[0](0,bc); mapsetbc[1](1,bc); }

/* specific soundstream */
typedef void (*_mapstream)(int,signed short*,int);
static void mapstream_nothing(int slot,signed short* stream,int len) { ; }
static _mapstream mapstream[2]={mapstream_nothing,mapstream_nothing};
void mapper_sound_stream(signed short* stream,int len) { mapstream[0](0,stream,len); mapstream[1](1,stream,len); }

/* specific cleanup */
typedef void (*_mapclean)(int);
static void mapclean_nothing(int slot) { ; }
static _mapclean mapclean[2]={mapclean_nothing,mapclean_nothing};

#include "mapper_table.h" /* in a separate file for readability.. only usable by mapper.c */

typedef struct {
	int uid;
	char* uid_fn;
	char* longname;
	char* shortname;
	void (*init)(int);
	u32 flags;
} _maptype;

static int sram_load(int,u8**,u32,const char*,u8);
static void sram_save(int,u8*,u32,const char*);
static void mapper_set_bios_default(void);
static __inline__ void refresh_cb(int);

const fp_mapread lut_mapread[2][8]={
	{ mapreadc10,	mapreadc12,	mapreadc14,	mapreadc16,	mapreadc18,	mapreadc1a,	mapreadc1c,	mapreadc1e },
	{ mapreadc20,	mapreadc22,	mapreadc24,	mapreadc26,	mapreadc28,	mapreadc2a,	mapreadc2c,	mapreadc2e }
};

#include "mapper_simple.c"
#include "mapper_misc.c"
#include "mapper_konami.c"
#include "mapper_ascii.c"
#include "mapper_panasonic.c"
#include "mapper_bootleg.c"

static const _maptype maptype[CARTTYPE_MAX]={
	/*uid	uid fn	long name			short name	init function		flags */
	{ 4,	"A8",	"ASCII 8",			"ASCII8",	mapinit_ascii8,		0 },
	{ 16,	"A8S",	"ASCII 8 + SRAM",		"ASCII8S",	mapinit_ascii8s,	MCF_SRAM },
	{ 5,	"A16",	"ASCII 16",			"ASCII16",	mapinit_ascii16,	0 },
	{ 15,	"A16S",	"ASCII 16 + SRAM",		"ASCII16S",	mapinit_ascii16s,	MCF_SRAM },
	{ 21,	"B80",	"Bootleg 80-in-1",		"BTL80in1",	mapinit_btl80,		0 },
	{ 22,	"B90",	"Bootleg 90-in-1",		"BTL90in1",	mapinit_btl90,		0 },
	{ 23,	"B126",	"Bootleg 126-in-1",		"BTL126in1",	mapinit_btl126,		0 },
	{ 9,	"CB",	"dB-Soft Cross Blaim",		"CBlaim",	mapinit_crossblaim,	0 },
	{ 6,	"RT",	"Irem R-Type",			"R-Type",	mapinit_rtype,		0 },
	{ 17,	"KOEI",	"Koei",				NULL,		mapinit_koei,		MCF_SRAM },
	{ 7,	"K",	"Konami",			NULL,		mapinit_konami,		0 },
	{ 19,	"GM2",	"Konami Game Master 2",		"GMaster2",	mapinit_gm2,		MCF_SRAM },
	{ 8,	"SCC",	"Konami SCC",			"SCC",		mapinit_konamiscc,	0 },
	{ 24,	"SCCI",	"Konami Sound Cartridge",	"SCC-I",	mapinit_konamiscci,	0 },
	{ 13,	"KS",	"Konami Synthesizer",		"KSyn",		mapinit_konamisyn,	0 },
	{ 20,	"MFR",	"Manuel Mega Flash ROM SCC",	"MFRSCC",	mapinit_mfrscc,		MCF_SRAM },
	{ 14,	"INK",	"Matra INK",			"INK",		mapinit_matraink,	0 },
	{ 10,	"HF",	"MicroCabin Harry Fox",		"HFox",		mapinit_harryfox,	0 },
	{ 18,	NULL,	"Panasonic PAC",		"PAC",		mapinit_pac,		MCF_SRAM|MCF_EMPTY },
	{ 0,	"M",	"Simple, mirrored",		"mirrored",	mapinit_mirror,		0 },
	{ 1,	"0",	"Simple, start at $0000",	"$0000",	mapinit_0000,		0 },
	{ 2,	"4",	"Simple, start at $4000",	"$4000",	mapinit_4000,		0 },
	{ 3,	"8",	"Simple, start at $8000",	"$8000",	mapinit_8000,		0 },
	{ 26,	"C",	"Simple, start at $C000",	"$C000",	mapinit_c000,		0 }, /* uid max */
	{ 12,	"PB",	"Sony Playball",		"Playball",	mapinit_playball,	0 },
	{ 11,	NULL,	"Vincent DSK2ROM",		"DSK2ROM",	mapinit_dsk2rom,	0 },
	{ 25,	"ZMB",	"Zemina ZMB",			"ZMB",		mapinit_zeminazmb,	0 }
};

int mapper_get_type_uid(u32 type)
{
	if (type>=CARTTYPE_MAX) type=CARTTYPE_MIRROR;
	return maptype[type].uid;
}

int mapper_get_uid_type(u32 uid)
{
	int i;
	if (uid>=CARTTYPE_MAX) uid=0;
	for (i=0;i<CARTTYPE_MAX;i++) if (maptype[i].uid==uid) return i;
	return CARTTYPE_MIRROR;
}

const char* mapper_get_type_longname(u32 type)
{
	if (type>=CARTTYPE_MAX) type=CARTTYPE_MIRROR;
	return maptype[type].longname;
}

const char* mapper_get_type_shortname(u32 type)
{
	if (type>=CARTTYPE_MAX) type=CARTTYPE_MIRROR;
	if (maptype[type].shortname) return maptype[type].shortname;
	else return maptype[type].longname;
}

u32 mapper_get_type_flags(u32 type)
{
	if (type>=CARTTYPE_MAX) type=CARTTYPE_MIRROR;
	return maptype[type].flags;
}

void mapper_set_carttype(int slot,u32 type) { carttype[slot&1]=type; }
int mapper_get_carttype(int slot) { return carttype[slot&1]; }
int mapper_get_cartsize(int slot) { return cartsize[slot&1]; }
u32 mapper_get_cartcrc(int slot) { return cartcrc[slot&1]; }
u32 mapper_get_bioscrc(void) { return bioscrc; }
int mapper_get_autodetect_crc(void) { return autodetect_crc; }
const char* mapper_get_file(int slot) { if (cartfile[slot&1]&&strlen(cartfile[slot&1])) return cartfile[slot&1]; else return NULL; }
const char* mapper_get_defaultbios_file(void) { return fn_default_bios; }
const char* mapper_get_bios_file(void) { return fn_bios; }
const char* mapper_get_bios_name(void) { return n_bios; }
void mapper_flush_ram(void) { memset(ram,RAMINIT,0x10000); }

char* mapper_get_current_name(char* c)
{
	MEM_CLEAN(file->name); /* set to NULL */
	if (mapper_get_file(1)) file_setfile(NULL,mapper_get_file(1),NULL,NULL);
	if (mapper_get_file(0)) file_setfile(NULL,mapper_get_file(0),NULL,NULL);
	
	if (file->name) sprintf(c,"%s",file->name);
	else {
		/* tape */
		if (tape_get_fn_cur()) file_setfile(NULL,tape_get_fn_cur(),NULL,NULL);
		if (file->name) sprintf(c,"%s",file->name);
		else {
			/* bios */
			if (fn_bios) file_setfile(NULL,fn_bios,NULL,NULL);
			if (file->name) sprintf(c,"%s",file->name);
			else sprintf(c,"none");
		}
	}
	
	return c;
}

u8 __fastcall mapreadnull(u16 a) { return UNMAPPED; }
void __fastcall mapwritenull(u16 a,u8 v) { ; }

u8 __fastcall mapreadbios(u16 a) { return bios[a]; }

/* only active when tape is inserted */
static int bh_tapion_offset=0;	static int bh_tapin_offset=0;	static int bh_tapoon_offset=0;	static int bh_tapout_offset=0;
u8 __fastcall mapreadbioshack(u16 a)
{
	if (bh_tapin_offset==a&&bh_tapin_offset==z80_get_pc()) { tape_read_byte(); return 0xc9; }
	else if (bh_tapout_offset==a&&bh_tapout_offset==z80_get_pc()) { tape_write_byte(); return 0xc9; }
	else if (bh_tapion_offset==a&&bh_tapion_offset==z80_get_pc()) { tape_read_header(); return 0xc9; }
	else if (bh_tapoon_offset==a&&bh_tapoon_offset==z80_get_pc()) { tape_write_header(); return 0xc9; }
	
	return bios[a];
}

u8 __fastcall mapreadram(u16 a) { return ram[a]; }
void __fastcall mapwriteram(u16 a,u8 v) { ram[a]=v; }

/* standard cart read */
u8 __fastcall mapreadc10(u16 a) { return cartmap[0][0][a&0x1fff]; }	u8 __fastcall mapreadc20(u16 a) { return cartmap[1][0][a&0x1fff]; }
u8 __fastcall mapreadc12(u16 a) { return cartmap[0][1][a&0x1fff]; }	u8 __fastcall mapreadc22(u16 a) { return cartmap[1][1][a&0x1fff]; }
u8 __fastcall mapreadc14(u16 a) { return cartmap[0][2][a&0x1fff]; }	u8 __fastcall mapreadc24(u16 a) { return cartmap[1][2][a&0x1fff]; }
u8 __fastcall mapreadc16(u16 a) { return cartmap[0][3][a&0x1fff]; }	u8 __fastcall mapreadc26(u16 a) { return cartmap[1][3][a&0x1fff]; }
u8 __fastcall mapreadc18(u16 a) { return cartmap[0][4][a&0x1fff]; }	u8 __fastcall mapreadc28(u16 a) { return cartmap[1][4][a&0x1fff]; }
u8 __fastcall mapreadc1a(u16 a) { return cartmap[0][5][a&0x1fff]; }	u8 __fastcall mapreadc2a(u16 a) { return cartmap[1][5][a&0x1fff]; }
u8 __fastcall mapreadc1c(u16 a) { return cartmap[0][6][a&0x1fff]; }	u8 __fastcall mapreadc2c(u16 a) { return cartmap[1][6][a&0x1fff]; }
u8 __fastcall mapreadc1e(u16 a) { return cartmap[0][7][a&0x1fff]; }	u8 __fastcall mapreadc2e(u16 a) { return cartmap[1][7][a&0x1fff]; }

static int mapper_patch_bios(u32 crc,u8* data)
{
	/* patch C-BIOS 0.21 PSG bug that's critical since meisei 1.2 (due to emulating the pin 6/7 quirk) */
	if (crc==0x1b3ab47f) {
		/* 2 instances of writing $80 to psg reg 15, change to $8f */
		data[0x152e]=data[0x15f1]=0x8f;
		return TRUE;
	}
	
	return FALSE;
}

void mapper_init(void)
{
	int i,j;
	
#if 0
	/* check mapper table for duplicates */
	u32 c; i=0;
	for (;;) {
		if (lutmaptype[i].type==CARTTYPE_MAX) break;
		c=lutmaptype[i].crc;
		
		j=0;
		for (;;) {
			if (lutmaptype[j].type==CARTTYPE_MAX) break;
			if (c==lutmaptype[j].crc&&i!=j) printf("dup %d %d - %08X\n",i,j,c);
			j++;
		}
		
		i++;
	}
#endif
	
	/* open default bios */
	file_setfile(&file->appdir,DEFAULT_BIOS,NULL,NULL);
	if (!file_open()||file->size!=0x8000||!file_read(default_bios,0x8000)) {
		file_close();
		LOG(LOG_MISC|LOG_ERROR,"Couldn't open default BIOS ROM.\nEnsure that %s\nis in the application directory.",DEFAULT_BIOS); exit(1);
	}
	default_bioscrc=file->crc32;
	MEM_CREATE(fn_default_bios,strlen(file->filename)+1);
	strcpy(fn_default_bios,file->filename);
	file_close();
	
	mapper_patch_bios(default_bioscrc,default_bios);
	
	memset(dummy,UNMAPPED,0x2000);
	
	autodetect_crc=TRUE; i=settings_get_yesnoauto(SETTINGS_CRCDETECTROM); if (i==FALSE||i==TRUE) autodetect_crc=i;
	
	/* get bios filename */
	n_bios=NULL; SETTINGS_GET_STRING(settings_info(SETTINGS_BIOS),&n_bios);
	if (n_bios==NULL||strlen(n_bios)==0) mapper_set_bios_default();
	else {
		file_setfile(&file->biosdir,n_bios,NULL,NULL);
		MEM_CREATE(fn_bios,strlen(file->filename)+1);
		strcpy(fn_bios,file->filename);
	}
	
	/* set standard handlers */
	for (j=0;j<4;j++)
	for (i=0;i<8;i++) {
		mapread_std[j][i]=mapreadnull;
		mapwrite_std[j][i]=mapwritenull;
	}
	/* slot 0: bios */
	for (i=0;i<8;i++) mapread_std[0][i]=mapreadbios;
	/* slot 3: ram */
	for (i=0;i<8;i++) {
		mapread_std[3][i]=mapreadram;
		mapwrite_std[3][i]=mapwriteram;
	}
	
	/* carts */
	for (j=0;j<2;j++)
	for (i=0;i<ROM_MAXPAGES;i++) {
		cart[j][i]=dummy;
	}
	cart[0][ROMPAGE_DUMMY]=cart[1][ROMPAGE_DUMMY]=dummy;
}

void mapper_clean(void)
{
	mapper_close_cart(0);
	mapper_close_cart(1);
	
	MEM_CLEAN(fn_default_bios);
	MEM_CLEAN(fn_bios); MEM_CLEAN(n_bios);
}


/* bios */
static void mapper_set_bios_default(void)
{
	memset(bios+0x8000,UNMAPPED,0x8000);
	memcpy(bios,default_bios,0x8000);
	bioscrc=default_bioscrc;
	
	MEM_CLEAN(n_bios);
	MEM_CREATE(n_bios,strlen(DEFAULT_BIOS)+1);
	sprintf(n_bios,DEFAULT_BIOS);
	
	MEM_CLEAN(fn_bios);
	MEM_CREATE(fn_bios,strlen(fn_default_bios)+1);
	strcpy(fn_bios,fn_default_bios);
	
	file_setfile(NULL,fn_bios,NULL,NULL);
	file_setdirectory(&file->biosdir);
}

void mapper_open_bios(const char* f)
{
	if (f==NULL) file_setfile(&file->biosdir,n_bios,NULL,"rom");
	else file_setfile(NULL,f,NULL,"rom");
	
	memset(bios,UNMAPPED,0x10000);
	
	if (!file_open()||file->size>0x10000||file->size<2||!file_read(bios,file->size)) {
		file_close();
		LOG(LOG_MISC|LOG_WARNING,"couldn't open BIOS, reverting to default\n");
		mapper_set_bios_default();
	}
	else {
		/* user specified */
		int len=strlen(file->name)+1;
		int ext=FALSE;
		if (file->ext&&strlen(file->ext)) {
			ext=TRUE;
			len+=strlen(file->ext)+1;
		}
		bioscrc=file->crc32;
		
		mapper_patch_bios(bioscrc,bios);
		
		MEM_CLEAN(n_bios);
		MEM_CREATE(n_bios,len);
		strcpy(n_bios,file->name);
		if (ext) {
			strcat(n_bios,".");
			strcat(n_bios,file->ext);
		}
		
		MEM_CLEAN(fn_bios);
		MEM_CREATE(fn_bios,strlen(file->filename)+1);
		strcpy(fn_bios,file->filename);
		
		file_close();
		file_setdirectory(&file->biosdir);
	}
}

void mapper_update_bios_hack(void)
{
	int i;
	
	if (tape_is_inserted()) {
		/* hacked */
		for (i=0;i<8;i++) mapread_std[0][i]=mapreadbioshack;
		
		/* set offsets */
		i=bios[0xe2]|bios[0xe3]<<8;
		if (bios[0xe1]==0xc3&&i<0x4000) bh_tapion_offset=i;
		else bh_tapion_offset=0xe1;
		
		i=bios[0xe5]|bios[0xe6]<<8;
		if (bios[0xe4]==0xc3&&i<0x4000) bh_tapin_offset=i;
		else bh_tapin_offset=0xe4;
		
		i=bios[0xeb]|bios[0xec]<<8;
		if (bios[0xea]==0xc3&&i<0x4000) bh_tapoon_offset=i;
		else bh_tapoon_offset=0xea;
		
		i=bios[0xee]|bios[0xef]<<8;
		if (bios[0xed]==0xc3&&i<0x4000) bh_tapout_offset=i;
		else bh_tapout_offset=0xed;
	}
	else {
		/* normal */
		for (i=0;i<8;i++) mapread_std[0][i]=mapreadbios;
	}
}


/* carts */
int mapper_open_cart(int slot,const char* f,int auto_patch)
{
	#define ROM_CLOSE(x) \
		file_patch_close(); file_close(); \
		if (!x) LOG(LOG_MISC|LOG_WARNING,"slot %d: couldn't open file\n",slot+1); \
		return x
	
	int i=TRUE;
	u32 crc,crc_p=0;
	char n[STRING_SIZE]={0};
	
	slot&=1;
	
	/* open file, just to get checksum before patching */
	file_setfile(NULL,f,NULL,"rom");
	if (!file_open()) { ROM_CLOSE(FALSE); }
	crc=file->crc32;
	file_close();
	
	if (auto_patch) {
		/* open optional ips (ones that enlarge filesize are unsupported in this case) */
		strcpy(n,file->name);
		file_setfile(&file->patchdir,n,"ips",NULL);
		if (!file_open()) {
			file_close();
			file_setfile(&file->patchdir,n,"zip","ips");
			if (!file_open()) { i=FALSE; file_close(); }
		}
		crc_p=file->crc32;
		if (i&(crc_p!=crc)) {
			LOG(LOG_MISC,"slot %d: applying patch (%sIPS)\n",slot+1,file->is_zip?"zipped ":"");
			if (!file_patch_init()) { file_patch_close(); crc_p=0; }
		}
		else crc_p=0;
		file_close();
		
		/* set to rom file again */
		file_setfile(NULL,f,NULL,"rom");
	}
	
	if (file_open()) {
		char* temp_fn;
		u32 temp_crc32;
		u32 temp_size;
		
		int pages=(file->size>>13)+((file->size&0x1fff)!=0);
		if (pages>ROM_MAXPAGES) { ROM_CLOSE(FALSE); }
		
		/* read rom */
		for (i=0;i<pages;i++) {
			MEM_CREATE(tempfile[i],0x2000); memset(tempfile[i],UNMAPPED,0x2000);
			if (!file_read(tempfile[i],0x2000)) {
				i++;
				while (i--) { MEM_CLEAN(tempfile[i]); }
				
				ROM_CLOSE(FALSE);
			}
		}
		
		/* success */
		MEM_CREATE(temp_fn,strlen(file->filename)+1);
		strcpy(temp_fn,file->filename);
		temp_crc32=file->crc32;
		temp_size=file->size;
		
		file_close();
		mapper_close_cart(slot); /* close previous */
		
		cartfile[slot]=temp_fn;
		cartcrc[slot]=temp_crc32;
		if (crc_p) cartcrc[slot]=(cartcrc[slot]<<1|(cartcrc[slot]>>31&1))^crc_p; /* combine with patch crc */
		
		/* set cart */
		for (i=0;i<pages;i++) cart[slot][i]=tempfile[i];
		for (cartmask[slot]=1;cartmask[slot]<pages;cartmask[slot]<<=1) { ; }
		cartmask[slot]--; cartsize[slot]=temp_size; cartpages[slot]=pages;
	}
	else { ROM_CLOSE(FALSE); }
	
	ROM_CLOSE(TRUE);
}

static void mapper_clean_custom(slot)
{
	mapclean[slot](slot);
	
	mapclean[slot]=mapclean_nothing;
	mapgetstatesize[slot]=mapgetstatesize_nothing;
	mapsavestate[slot]=mapsavestate_nothing;
	maploadstatecur[slot]=maploadstatecur_nothing;
	maploadstate[slot]=maploadstate_nothing;
	mapsetbc[slot]=mapsetbc_nothing;
	mapstream[slot]=mapstream_nothing;
	mapnf[slot]=mapnf_nothing;
	mapef[slot]=mapef_nothing;
}

void mapper_close_cart(int slot)
{
	int i;
	
	slot&=1;
	mapper_clean_custom(slot);
	
	for (i=0;i<ROM_MAXPAGES;i++) {
		if (cart[slot][i]!=dummy) {
			MEM_CLEAN(cart[slot][i]);
			cart[slot][i]=dummy;
		}
	}
	
	MEM_CLEAN(cartfile[slot]);
	cartcrc[slot]=cartpages[slot]=cartmask[slot]=cartsize[slot]=cartsramsize[slot]=0;
	cartsram[slot]=NULL;
}

u32 mapper_autodetect_cart(const char* f)
{
	int i=0;
	u8 filedata[0x10000];
	u32 ret=CARTTYPE_MIRROR;
	
	file_setfile(NULL,f,NULL,"rom");
	if (!file_open()||file->size<0x2000||file->size>ROM_MAXSIZE||!file_read(filedata,file->size<0x10000?file->size:0x10000)) {
		file_close(); return CARTTYPE_MIRROR;
	}
	file_close();
	
	/* try parse, {x} in filename where x=id */
	if (file->name&&strlen(file->name)) {
		int begin,end,len=strlen(file->name);
		
		for (begin=0;begin<len;begin++) if (file->name[begin]=='{') break;
		for (end=begin;end<len;end++) if (file->name[end]=='}') break;
		begin++;
		
		if (begin<end&&end<len&&(end-begin)<0x10) {
			char uid_fn[0x10];
			memset(uid_fn,0,0x10);
			memcpy(uid_fn,file->name+begin,end-begin);
			
			for (i=0;i<CARTTYPE_MAX;i++) if (maptype[i].uid_fn&&strlen(maptype[i].uid_fn)&&stricmp(uid_fn,maptype[i].uid_fn)==0) return i;
		}
	}
	
	/* try lookuptable */
	if (autodetect_crc) {
		u32 crc=file->crc32;
		while (lutmaptype[i].type!=CARTTYPE_MAX) {
			if (lutmaptype[i].crc==crc) return lutmaptype[i].type;
			i++;
		}
	}
	
	/* guess type */
	switch (file->size) {
		/* 8KB/16KB */
		case 0x2000: case 0x4000: {
			u16 start=filedata[3]<<8|filedata[2];
			
			if (start==0) {
				/* start address of $0000: call address in the $4000 region: $4000, else $8000 */
				u16 call=filedata[5]<<8|filedata[4];
				if (call>0x4000&&call<0x8000) ret=CARTTYPE_START4000;
				else ret=CARTTYPE_START8000;
			}
			/* start address in the $8000 region: $8000 */
			else if (start>0x8000&&start<0xc000) ret=CARTTYPE_START8000;
			else ret=CARTTYPE_MIRROR;
			
			break;
		}
		
		/* 32KB */
		case 0x8000:
			/* no "AB" at $0000, but "AB" at $4000 */
			if (filedata[0]!='A'&&filedata[1]!='B'&&filedata[0x4000]=='A'&&filedata[0x4001]=='B') {
				u16 start=filedata[0x4003]<<8|filedata[0x4002];
				
				/* start address of 0 or in $8000 region: BASIC type */
				if (start==0||(start>0x8000&&start<0xc000)) ret=CARTTYPE_MIRROR;
				else ret=CARTTYPE_START0000;
			}
			else ret=CARTTYPE_MIRROR;
			
			break;
		
		/* 48KB */
		case 0xc000:
			/* start with "AB" (but not at offset $4000): $4000, not "AB": $0000 */
			if (filedata[0]=='A'&&filedata[1]=='B'&&filedata[0x4000]!='A'&&filedata[0x4001]!='B') ret=CARTTYPE_START4000;
			else ret=CARTTYPE_START0000;
			
			break;
		
		/* 64KB/128KB/256KB/512KB */
		case 0x10000: case 0x20000: case 0x40000: case 0x80000:
			/* 64KB and no "AB" at start: $0000 */
			if (file->size==0x10000&&filedata[0]!='A'&&filedata[1]!='B') ret=CARTTYPE_START0000;
			
			/* (I could autodetect by searching for common bankswitch address writes, like openMSX did/does,
			 but that would be inaccurate) */
			/* autodetect all other megaROMs as ASCII8 */
			else ret=CARTTYPE_ASCII8;
			
			break;
		
		/* 384KB, only type with that size is R-Type */
		case 0x60000: ret=CARTTYPE_RTYPE; break;
		
		/* disk sizes (360KB and 720KB are most common) */
		case 163840: case 184320: case 327680: case 368640:
		case 655360: case 737280: case 1474560:
			ret=CARTTYPE_DSK2ROM;
			break;
		
		/* other */
		default:
			if (file->size>0x10000) ret=CARTTYPE_ASCII8;
			else ret=CARTTYPE_MIRROR;
			break;
	}
	
	return ret;
}

void mapper_init_cart(int slot)
{
	int i;
	
	slot&=1;
	mapper_clean_custom(slot);
	
	for (i=0;i<8;i++) {
		/* default read handlers */
		mapread_std[slot+1][i]=lut_mapread[slot][i];
		
		/* default write handlers */
		mapwrite_std[slot+1][i]=mapwritenull;
		
		/* empty memmap */
		cartbank[slot][i]=ROMPAGE_DUMMY;
		cartmap[slot][i]=dummy;
	}
	
	maptype[carttype[slot]].init(slot);
	refresh_cb(slot);
}

static __inline__ void refresh_cb(int slot)
{
	/* refresh cart banks */
	cartmap[slot][0]=cart[slot][cartbank[slot][0]];
	cartmap[slot][1]=cart[slot][cartbank[slot][1]];
	cartmap[slot][2]=cart[slot][cartbank[slot][2]];
	cartmap[slot][3]=cart[slot][cartbank[slot][3]];
	cartmap[slot][4]=cart[slot][cartbank[slot][4]];
	cartmap[slot][5]=cart[slot][cartbank[slot][5]];
	cartmap[slot][6]=cart[slot][cartbank[slot][6]];
	cartmap[slot][7]=cart[slot][cartbank[slot][7]];
}

void __fastcall mapper_refresh_cb(int slot) { refresh_cb(slot&1); }


/* primary slot register */
void __fastcall mapper_refresh_pslot_read(void)
{
	mapread[0]=mapread_std[pslot&3][0];		mapread[1]=mapread_std[pslot&3][1];
	mapread[2]=mapread_std[pslot>>2&3][2];		mapread[3]=mapread_std[pslot>>2&3][3];
	mapread[4]=mapread_std[pslot>>4&3][4];		mapread[5]=mapread_std[pslot>>4&3][5];
	mapread[6]=mapread_std[pslot>>6&3][6];		mapread[7]=mapread_std[pslot>>6&3][7];
}

void __fastcall mapper_refresh_pslot_write(void)
{
	mapwrite[0]=mapwrite_std[pslot&3][0];		mapwrite[1]=mapwrite_std[pslot&3][1];
	mapwrite[2]=mapwrite_std[pslot>>2&3][2];	mapwrite[3]=mapwrite_std[pslot>>2&3][3];
	mapwrite[4]=mapwrite_std[pslot>>4&3][4];	mapwrite[5]=mapwrite_std[pslot>>4&3][5];
	mapwrite[6]=mapwrite_std[pslot>>6&3][6];	mapwrite[7]=mapwrite_std[pslot>>6&3][7];
}

void __fastcall mapper_write_pslot(u8 v)
{
	pslot=v;
	
	mapper_refresh_pslot_read();
	mapper_refresh_pslot_write();
}

u8 __fastcall mapper_read_pslot(void) { return pslot; }


/* SRAM */
static int sram_netplay[2]={0,0};

static int sram_load(int slot,u8** d,u32 size,const char* fnc,u8 fill)
{
	int ret=FALSE;
	char fn[STRING_SIZE];
	
	if (size==0) return FALSE;
	
	MEM_CREATE_N(*d,size);
	memset(*d,fill,size);
	cartsram[slot]=*d; cartsramsize[slot]=size;
	
	if (!fnc&&!mapper_get_file(slot)) return FALSE; /* empty slot */
	
	if (netplay_is_active()) {
		/* don't load (or save) if netplay */
		LOG(LOG_MISC|LOG_WARNING|LOG_TYPE(LT_NETPLAYSRAM),"SRAM is volatile during netplay");
		sram_netplay[slot]=TRUE;
		return FALSE;
	}
	
	if (fnc) file_setfile(&file->batterydir,fnc,NULL,NULL);
	else {
		file_setfile(NULL,mapper_get_file(slot),NULL,NULL);
		strcpy(fn,file->name);
		file_setfile(&file->batterydir,fn,"sav",NULL);
	}
	
	if (!file_accessible()) {
		/* SRAM file didn't exist yet */
		if (!file_save()||!file_write(*d,size)) LOG(LOG_MISC|LOG_WARNING,"couldn't init slot %d battery backed SRAM\n",slot+1);
	}
	
	/* load */
	else if (!file_open()||file->size!=size||!file_read(*d,size)) {
		LOG(LOG_MISC|LOG_WARNING,"couldn't load slot %d battery backed SRAM\n",slot+1);
		memset(*d,fill,size);
	}
	
	else ret=TRUE;
	
	file_close();
	
	MEM_CREATE(cartsramcopy[slot],size);
	memcpy(cartsramcopy[slot],*d,size);
	
	return ret;
}

static void sram_save(int slot,u8* d,u32 size,const char* fnc)
{
	int err=FALSE;
	char fn[STRING_SIZE];
	
	if (sram_netplay[slot]||d==NULL||size==0||(!fnc&&!mapper_get_file(slot))||(cartsramcopy[slot]&&memcmp(cartsramcopy[slot],d,size)==0)) err=TRUE;
	
	sram_netplay[slot]=FALSE;
	MEM_CLEAN(cartsramcopy[slot]);
	
	if (err) { MEM_CLEAN(d); return; }
	
	if (fnc) file_setfile(&file->batterydir,fnc,NULL,NULL);
	else {
		file_setfile(NULL,mapper_get_file(slot),NULL,NULL);
		strcpy(fn,file->name);
		file_setfile(&file->batterydir,fn,"sav",NULL);
	}
	
	/* save */
	if (!file_save()||!file_write(d,size)) LOG(LOG_MISC|LOG_WARNING,"couldn't save slot %d battery backed SRAM\n",slot+1);
	file_close();
	
	/* clean */
	MEM_CLEAN(d);
}


/* state		size
ram			0x10000
pslot			1

cart banknums		2*2*8 (32)

==			0x10000+33

type specific		?

*/
/* version derivatives: mapper*.c, am29f040b.c, sample.c, scc.c */
#define STATE_VERSION	2
/* version history:
1: initial
2: SRAM crash bugfix + changes in scc.c
*/
#define STATE_SIZE	(0x10000+33)

int __fastcall mapper_state_get_version(void)
{
	return STATE_VERSION;
}

int __fastcall mapper_state_get_size(void)
{
	return STATE_SIZE+cartsramsize[0]+cartsramsize[1]+mapgetstatesize[0](0)+mapgetstatesize[1](1);
}

void __fastcall mapper_state_save(u8** s)
{
	STATE_SAVE_C(ram,0x10000);
	STATE_SAVE_1(pslot);
	
	STATE_SAVE_2(cartbank[0][0]);	STATE_SAVE_2(cartbank[0][1]);	STATE_SAVE_2(cartbank[0][2]);	STATE_SAVE_2(cartbank[0][3]);
	STATE_SAVE_2(cartbank[0][4]);	STATE_SAVE_2(cartbank[0][5]);	STATE_SAVE_2(cartbank[0][6]);	STATE_SAVE_2(cartbank[0][7]);
	STATE_SAVE_2(cartbank[1][0]);	STATE_SAVE_2(cartbank[1][1]);	STATE_SAVE_2(cartbank[1][2]);	STATE_SAVE_2(cartbank[1][3]);
	STATE_SAVE_2(cartbank[1][4]);	STATE_SAVE_2(cartbank[1][5]);	STATE_SAVE_2(cartbank[1][6]);	STATE_SAVE_2(cartbank[1][7]);
	
	if (cartsramsize[0]) { STATE_SAVE_C(cartsram[0],cartsramsize[0]); }
	if (cartsramsize[1]) { STATE_SAVE_C(cartsram[1],cartsramsize[1]); }
	
	mapsavestate[0](0,s);
	mapsavestate[1](1,s);
}

void __fastcall mapper_state_load_cur_inc(u8** s)
{
	STATE_LOAD_C(ram,0x10000);
	STATE_LOAD_1(pslot);
	
	STATE_LOAD_2(cartbank[0][0]);	STATE_LOAD_2(cartbank[0][1]);	STATE_LOAD_2(cartbank[0][2]);	STATE_LOAD_2(cartbank[0][3]);
	STATE_LOAD_2(cartbank[0][4]);	STATE_LOAD_2(cartbank[0][5]);	STATE_LOAD_2(cartbank[0][6]);	STATE_LOAD_2(cartbank[0][7]);
	STATE_LOAD_2(cartbank[1][0]);	STATE_LOAD_2(cartbank[1][1]);	STATE_LOAD_2(cartbank[1][2]);	STATE_LOAD_2(cartbank[1][3]);
	STATE_LOAD_2(cartbank[1][4]);	STATE_LOAD_2(cartbank[1][5]);	STATE_LOAD_2(cartbank[1][6]);	STATE_LOAD_2(cartbank[1][7]);
	
	if (cartsramsize[0]) { STATE_LOAD_C(cartsram[0],cartsramsize[0]); }
	if (cartsramsize[1]) { STATE_LOAD_C(cartsram[1],cartsramsize[1]); }
}

void __fastcall mapper_state_load_cur(u8** s)
{
	mapper_state_load_cur_inc(s);
	
	maploadstatecur[0](0,s);
	maploadstatecur[1](1,s);
}

int __fastcall mapper_state_load(int v,u8** s)
{
	switch (v) {
		case 1:
			mapper_state_load_cur_inc(s);
			maploadstate[0](0,v,s);
			maploadstate[1](1,v,s);
			break;
		case STATE_VERSION:
			mapper_state_load_cur(s);
			break;
		
		default: return FALSE;
	}
	
	return TRUE;
}
